# core
import json
from dataclasses import asdict

# lib
import requests
from requests.compat import urljoin

# files
from .settings import YALIDINE_URL
from .entity import (
    Parcel,
    ParcelFilter,
    HistoryFilter,
    CenterFilter,
    CommunesFilter,
    WilayasFilter,
    DeliveryFeesFilter,
)


exclude_none_value = lambda x: {k: v for (k, v) in x if v is not None}

asdict_true_value = lambda x: asdict(x, dict_factory=exclude_none_value)

dict_to_query_string = lambda x: "&".join([f"{k}={v}" for k, v in x.items()])


def response_or_exception(fn):
    from functools import wraps

    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        response: requests.Response = fn(self, *args, **kwargs)
        if response.status_code == 422:
            raise requests.exceptions.HTTPError(response, response=response)
        response.raise_for_status()
        self.last_response_headers = response.headers
        return response.json()

    return wrapper


# ===============================
# MARK: YalidineClient
# ===============================
class YalidineClient:
    def __init__(self, api_id, api_token, url=YALIDINE_URL):
        self.url = url
        self.api_id = api_id
        self.api_token = api_token
        self.last_response_headers = None
        self.headers = {
            "X-API-ID": self.api_id,
            "X-API-TOKEN": self.api_token,
            "Content-Type": "application/json",
        }

    # ===============================
    # MARK: Parcels
    # ===============================
    @response_or_exception
    def get_parcel(self, parcel_id):
        response = requests.get(
            urljoin(self.url, f"parcels/{parcel_id}"),
            headers=self.headers,
        )

        return response

    @response_or_exception
    def get_parcels(self, filter: ParcelFilter = None):
        query = ""

        if filter:
            f = asdict_true_value(filter)
            query = dict_to_query_string(f)

        response = requests.get(
            urljoin(self.url, f"parcels/?{query}"),
            headers=self.headers,
        )
        print(response.url)
        return response

    @response_or_exception
    def create_parcel(self, parcel_list: list[Parcel]):
        assert isinstance(parcel_list, list)
        
        body = [asdict(parcel) for parcel in parcel_list]

        response = requests.post(
            urljoin(self.url, "parcels"),
            headers=self.headers,
            json=body,
        )

        return response

    @response_or_exception
    def update_parcel(self, parcel_id: str, data: Parcel):
        response = requests.patch(
            urljoin(self.url, f"parcels/{parcel_id}"),
            headers=self.headers,
            json=asdict_true_value(data),
        )

        return response

    @response_or_exception
    def delete_parcel(self, parcel_id: str):
        response = requests.delete(
            urljoin(self.url, f"parcels/{parcel_id}"),
            headers=self.headers,
        )

        return response

    @response_or_exception
    def delete_parcels(self, parcel_ids: list[str]):
        response = requests.delete(
            urljoin(self.url, f"parcels/?tracking={','.join(parcel_ids)}"),
            headers=self.headers,
        )
        return response

    # ===============================
    # MARK: Histories
    # ===============================

    @response_or_exception
    def get_histories(self, filter: HistoryFilter = None):
        query = ""

        if filter:
            f = asdict_true_value(filter)
            query = dict_to_query_string(f)

        response = requests.get(
            urljoin(self.url, f"histories/?{query}"),
            headers=self.headers,
        )
        print(response.url)

        return response

    @response_or_exception
    def get_parcel_history(self, traking_id: str):
        response = requests.get(
            urljoin(self.url, f"histories/{traking_id}"),
            headers=self.headers,
        )

        return response

    # ===============================
    # MARK: Centers
    # ===============================
    @response_or_exception
    def get_centers(self, filter: CenterFilter = None):
        query = ""

        if filter:
            f = asdict_true_value(filter)
            query = dict_to_query_string(f)

        response = requests.get(
            urljoin(self.url, f"centers/?{query}"),
            headers=self.headers,
        )

        return response

    @response_or_exception
    def get_center(self, center_id: str):
        response = requests.get(
            urljoin(self.url, f"centers/{center_id}"),
            headers=self.headers,
        )

        return response

    # ===============================
    # MARK: Communes
    # ===============================
    @response_or_exception
    def get_communes(self, filter: CommunesFilter = None):
        query = ""

        if filter:
            f = asdict_true_value(filter)
            query = dict_to_query_string(f)

        response = requests.get(
            urljoin(self.url, f"communes/?{query}"),
            headers=self.headers,
        )

        return response

    @response_or_exception
    def get_commune(self, commune_id: str):
        response = requests.get(
            urljoin(self.url, f"communes/{commune_id}"),
            headers=self.headers,
        )

        return response

    # ===============================
    # MARK: Wilayas
    # ===============================
    @response_or_exception
    def get_wilayas(self, filter: WilayasFilter = None):
        query = ""

        if filter:
            f = asdict_true_value(filter)
            query = dict_to_query_string(f)

        response = requests.get(
            urljoin(self.url, f"wilayas/?{query}"),
            headers=self.headers,
        )

        return response

    @response_or_exception
    def get_wilaya(self, wilaya_id: str):
        response = requests.get(
            urljoin(self.url, f"wilayas/{wilaya_id}"),
            headers=self.headers,
        )

        return response

    # ===============================
    # MARK: Delivery Fees
    # ===============================
    @response_or_exception
    def get_delivery_fees(self, filter: DeliveryFeesFilter = None):
        query = ""

        if filter:
            f = asdict_true_value(filter)
            query = dict_to_query_string(f)

        response = requests.get(
            urljoin(self.url, f"deliveryfees/?{query}"),
            headers=self.headers,
        )

        return response

    @response_or_exception
    def get_wilaya_delivery_fee(self, wilaya_id: str):
        response = requests.get(
            urljoin(self.url, f"deliveryfees/{wilaya_id}"),
            headers=self.headers,
        )

        return response



import os
import unittest

from dotenv import load_dotenv
import requests

from src.yalidine.api import YalidineClient
from src.yalidine.entity import (
    Parcel,
    ParcelFilter,
    HistoryFilter,
    CenterFilter,
    WilayasFilter,
    CommunesFilter,
    DeliveryFeesFilter,
)


class TestYalidineAPI(unittest.TestCase):
    def setUp(self) -> None:
        load_dotenv()
        self.api_id = os.getenv("API_ID")
        self.api_token = os.getenv("API_TOKEN")
        self.client = YalidineClient(self.api_id, self.api_token)

    #===============================
    # MARK: Test Limits status
    #===============================
    def test_get_limits(self):
        response = self.client.get_parcels()
        limit = self.client.last_response_headers
        print(limit)
        self.assertIsInstance(limit, dict)

    # ===============================
    # MARK: Parcels
    # ===============================

    def test_get_parcels(self):
        response = self.client.get_parcels()
        self.assertIsInstance(response, dict)

    def test_crud_parcel(self):
        parcel = Parcel(
            order_id="test_parcel",
            firstname="firstname",
            familyname="familyname",
            contact_phone="0555885588",
            address="Bouira",
            from_wilaya_name="Bouira",
            to_commune_name="Aïn Bessem",
            to_wilaya_name="Bouira",
            product_list="Press a cafe",
            price=3000,
            do_insurance=False,
            declared_value=3500,
            height=10,
            width=10,
            length=30,
            weight=10,
            freeshipping=False,
            is_stopdesk=False,
            has_exchange=0,
            product_to_collect="Test",
        )

        response = self.client.create_parcel([parcel])
        self.assertIsNotNone(response["test_parcel"])

        parcel_id = response["test_parcel"]["tracking"]

        parcel.familyname = "familyname_2"
        response = self.client.update_parcel(parcel_id=parcel_id, data=parcel)

        self.assertIsInstance(response, dict)
        self.assertEqual(response['familyname'],'familyname_2')

        response = self.client.get_parcel(parcel_id=parcel_id)
        self.assertIsInstance(response, dict)
        self.assertIsInstance(response['data'],list)
        self.assertEqual(response['data'][0]['familyname'],'familyname_2')

        response = self.client.delete_parcel(parcel_id=parcel_id)
        self.assertTrue(response[0]["deleted"])

    # ===============================
    # MARK: History
    # ===============================
    def test_get_histories(self):
        response = self.client.get_histories(filter=HistoryFilter(page_size=2))
        print(response)
        self.assertIsInstance(response, dict)

    #===============================
    # MARK: Centers
    #===============================
    def test_get_centers(self):
        response = self.client.get_centers(filter=CenterFilter(page_size=2))
        print(response)
        self.assertIsInstance(response, dict)

    # ===============================
    # MARK: Wilayas
    # ===============================
    def test_get_wilayas(self):
        response = self.client.get_wilayas(filter=WilayasFilter(page_size=2))
        print(response)
        self.assertIsInstance(response, dict)

    # ===============================
    # MARK: Communes
    # ===============================

    def test_get_communes(self):
        response = self.client.get_communes(filter=CommunesFilter(wilaya_id="10",page_size=2))
        print(response)
        self.assertIsInstance(response, dict)

    # ===============================
    # MARK: Delivery Fees
    # ===============================
    def test_get_delivery_fees(self):
        response = self.client.get_delivery_fees(
            filter=DeliveryFeesFilter(wilaya_id="10")
        )
        print(response)
        self.assertIsInstance(response, dict)



from dataclasses import dataclass, field
from typing import Optional, List




# ===============================
# MARK: Parcel
# ===============================
@dataclass
class Parcel:
    """
	Parcel object

	...

	Attributes
	----------
		order_id : str
		A string representing the order id of the parcel
		from_wilaya_name : str
			A string representing the sender's wilaya name.
		firstname : str
			The receiver's first name.
		familyname : str
			The receiver's family name.
		contact_phone : str
			The receiver's phone numbers (can be separated by commas if many)
		address : str
			The receiver's address
		to_commune_name : str
			A string representing the receiver's commune name.
		to_wilaya_name : str
			A string representing the receiver's wilaya name.
		product_list : str
			The description of the shipment's content.
		price : int
			An integer amount representing the price you want to recover from the receiver. (equal or between 0 and 150000).
		do_insurance : bool
			Whether or not you opt for an insurance. if `True`, 0% fee of `declared_value` is applicable, the refund is 100%.
		declared_value : int
			Represents the financial estimation of the items within the parcel. (must be between 0 and 150000).
		length : int
			An integer amount representing the length of the parcel's content. (greater than or equal to 0).
		width : int
			An integer amount representing the width of the parcel's content. (greater than or equal to 0).
		height : int
			An integer amount representing the height of the parcel's content. (greater than or equal to 0).
		weight : int
			An integer amount representing the weight of the parcel's content. (greater than or equal to 0).
		freeshipping : bool
			A Boolean representing whether the delivery fee is free (paid by the sender) or not.
				- `True = paid by the sender`.
				- `False = paid by the receiver`.
		is_stopdesk : bool
			Whether the delivery will be done in a stop-desk or home delivery.
				- `True = delivery` in stop desk, you must include the param `stopdesk_id`.
				- `False = home delivery`.
		stopdesk_id : str
			This parameter is required if `is_stopdesk` is true, optional if not.
		has_exchange : bool
			A Boolean representing Whether or not you want to make an exchange request for this parcel.
		product_to_collect: str
			This parameter is required if `has_exchange` is true, optional if not.

    """

    order_id: Optional[str] = None
    from_wilaya_name: Optional[str] = None
    firstname: Optional[str] = None
    familyname: Optional[str] = None
    contact_phone: Optional[str] = None
    address: Optional[str] = None
    to_commune_name: Optional[str] = None
    to_wilaya_name: Optional[str] = None
    product_list: Optional[str] = None
    price: Optional[int] = None
    do_insurance: Optional[bool] = None
    declared_value: Optional[int] = None
    length: Optional[int] = None
    width: Optional[int] = None
    height: Optional[int] = None
    weight: Optional[int] = None
    freeshipping: Optional[bool] = None
    is_stopdesk: Optional[bool] = None
    stopdesk_id: Optional[str] = None
    has_exchange: Optional[bool] = None
    product_to_collect: Optional[str] = None


# ===============================
# MARK: ParcelFilter
# ===============================
@dataclass
class ParcelFilter:
	"""
	ParcelFilter object

	...

	Attributes
	----------
		tracking : str
			The unique identifier for the parcel.
				- eg: one item \"yal-123456\"
				- eg: list of items \"yal-123456,yal-789123,yal-456789\"
		order_id: str
			The receiver’s order id
		import_id: int
			The id of the operation of bulk-creation of the parcel (through importation or API creation)
		to_wilaya_id: int
			The id of the receiver’s wilaya
		to_commune_name: str
			The receiver’s commune name.
		is_stopdesk: bool
			Whether the delivery is done in a stop-desk or home delivery.
				- `True` means delivery in stop desk
				- `False` means home delivery
		is_exchange: bool
			Whether or not the package is the annexed parcel for an exchange request.
		has_exchange: bool
			Whether or not you want to make an exchange request for this parcel.
		freeshipping: bool
			Whether the delivery fee is free (paid by the sender).
		date_creation: str
			The parcel’s date of creation in the format **YYYY-MM-DD**
				- Only one date: `date_creation=2021-01-01`
				- Between two dates: `date_creation=2021-01-01,2021-01-31`
        date_last_status: str
			The parcel’s date of the last status in the format **YYYY-MM-DD**
				- Only one date: `date_last_status=2021-01-01`
				- Between two dates: `date_last_status=2021-01-01,2021-01-31`
		payment_status: str
			The current payment status of the parcel.
				- not-ready
				- ready
				- receivable
				- payed
        last_status: str
			The current status of the parcel's delivery.
				- Pas encore expédié
				- A vérifier
				- En préparation
				- Pas encore ramassé
				- Prêt à expédier
				- Ramassé
				- Bloqué
				- Débloqué
				- Transfert
				- Expédié
				- Centre
				- En localisation
				- Vers Wilaya
				- Reçu à Wilaya
				- En attente du client
				- Sorti en livraison
				- En attente
				- En alerte
				- Tentative échouée
				- Livré
				- Echèc livraison
				- Retour vers centre
				- Retourné au centre
				- Retour transfert
				- Retour groupé
				- Retour à retirer
				- Retour vers vendeur
				- Retourné au vendeur
				- Echange échoué
        fields: str
			You can specify which fields you want returned by using the field parameter and listing each field separated by a comma.
        
		page : int
			The number of the page you would request
		page_size : int
			A limit on the number of objects to be returned, between 1 and 1000
               
        order_by: str
				- date_creation
				- date_last_status 
				- tracking 
				- order_id 
				- import_id 
				- to_wilaya_id 
				- to_commune_id 
				- last_status
                    
        desc: bool
			Order the result descening
        asc: bool
			Order the result ascending
	"""
		
	tracking: Optional[str] = None
	order_id: Optional[str] = None
	import_id: Optional[int] = None
	to_wilaya_id: Optional[int] = None
	to_commune_id: Optional[int] = None
	is_stopdesk: Optional[bool] = None
	is_exchange: Optional[bool] = None
	has_exchange: Optional[bool] = None
	freeshipping: Optional[bool] = None
	date_creation: Optional[str] = None
	date_last_status: Optional[str] = None
	payment_status: Optional[str] = None
	last_status: Optional[str] = None
	fields: Optional[str] = None
	page: Optional[int] = None
	page_size: Optional[int] = None
	order_by: Optional[str] = None
	desc: Optional[bool] = None
	asc: Optional[bool] = None


#===============================
# MARK:  HistoryFilter
#===============================


@dataclass
class HistoryFilter:
    """
	HistoryFilter object

	...

	Attributes
	----------
		tracking : str
			The identifier of the parcels.
		status: str
			The status of the parcel
				- Pas encore expédié
				- A vérifier
				- En préparation
				- Pas encore ramassé
				- Prêt à expédier
				- Ramassé
				- Bloqué
				- Débloqué
				- Transfert
				- Expédié
				- Centre
				- En localisation
				- Vers Wilaya
				- Reçu à Wilaya
				- En attente du client
				- Sorti en livraison
				- En attente
				- En alerte
				- Alerte résolue
				- Tentative échouée
				- Livré
				- Echèc livraison
				- Retour vers centre
				- Retourné au centre
				- Retour transfert
				- Retour groupé
				- Retour à retirer
				- Retour vers vendeur
				- Retourné au vendeur
				- Echange échoué
        date_status: str
			The status’s date of creation in the format YYYY-MM-DD.
				- Only one date: `date_status=2021-01-01`
				- Between two dates: `date_status=2021-01-01,2021-01-31`
        reason: str
			The reason of a failed delivery attempt or a parcel hold
		fields: str
			You can specify which fields you want returned by using the field parameter and listing each field separated by a comma.	
		page : int
			the number of the page you would request
		page_size : int
			the number of result in the same page
		order_by: str
			By default, the results in the histories response are ordered by the date_status in a descending way.
		desc: bool
			Order the result descening
        asc: bool
			Order the result ascending

	"""
    tracking: Optional[str] = None
    status: Optional[str] = None
    date_status: Optional[str] = None
    reason: Optional[str] = None
    fields: Optional[str] = None
    page: Optional[int] = None
    page_size: Optional[int] = None
    order_by: Optional[str] = None
    desc: Optional[bool] = None
    asc: Optional[bool] = None
    

#===============================
# MARK:  CenterFilter
#===============================

@dataclass
class CenterFilter:
	"""
	CenterFilter object

	...

	Attributes
	----------
		center_id: str
            The identifier of the center
		commune_id: str
            The identifier of the center's commune
		commune_name: str
            The commune’s name of the center.
		wilaya_id: str
            The identifier of the center's wilaya
		wilaya_name: str
            The wilaya’s name of the center.
		fields: str
            You can specify which fields you want returned by using the field parameter and listing each field separated by a comma.
        page : int
			the number of the page you would request
		page_size : int
			the number of result in the same page
		order_by: str
			By default, the result in the centers results is ordered by center_id in an ascending way
        desc: bool
			Order the result descening
        asc: bool
			Order the result ascending
	"""
	center_id: Optional[str] = None
	commune_id: Optional[str] = None
	commune_name: Optional[str] = None
	wilaya_id: Optional[str] = None
	wilaya_name: Optional[str] = None
	fields: Optional[str] = None
	page: Optional[int] = None
	page_size: Optional[int] = None
	order_by: Optional[str] = None
	desc: Optional[bool] = None
	asc: Optional[bool] = None
      
#===============================
# MARK: CommunesFilter
#===============================

@dataclass
class CommunesFilter:
	"""
	CommunesFilter object

	...

	Attributes
	----------
		id: str
			The identifier of the commune
		wilaya_id: str
			The commune’s wilaya id
		has_stop_desk: bool
			Whether or not the commune has a stop desk
		is_deliverable: bool
			Whether or not the commune is deliverable
		fields: str
			You can specify which fields you want returned by using the field parameter and listing each field separated by a comma.
		page : int
			the number of the page you would request
		page_size : int
			the number of result in the same page
		order_by: str
			By default, the result in the communes results is ordered by id in an ascending way
		desc: bool
			Order the result descening
		asc: bool
			Order the result ascending
	"""
	id: Optional[str] = None
	wilaya_id: Optional[str] = None
	has_stop_desk: Optional[bool] = None
	is_deliverable: Optional[bool] = None
	fields: Optional[str] = None
	page: Optional[int] = None
	page_size: Optional[int] = None
	order_by: Optional[str] = None
	desc: Optional[bool] = None
	asce: Optional[bool] = None

#===============================
# MARK: WilayasFilter
#===============================

@dataclass
class WilayasFilter:
	"""
	WilayasFilter object

	...

	Attributes
		id: str
			The identifier of the wilaya
		name: str
			The wilaya’s name
		fields: str
			You can specify which fields you want returned by using the field parameter and listing each field separated by a comma.
		page : int
			the number of the page you would request
		page_size : int
			the number of result in the same page
		order_by: str
			By default, the result in the wilayas results is ordered by id in an ascending way
		desc: bool
			Order the result descening
		asc: bool
			Order the result ascending
	"""
	id: Optional[str] = None
	name: Optional[str] = None
	fields: Optional[str] = None
	page: Optional[int] = None
	page_size: Optional[int] = None
	order_by: Optional[str] = None
	desc: Optional[bool] = None
	asc: Optional[bool] = None

#===============================
# MARK: DeliveryFeesFilter
#===============================

@dataclass
class DeliveryFeesFilter:
	"""
	DeliveryFeesFilter object

	...

	Attributes
		wilaya_id: str
			The wilaya’s id
		fields: str
			You can specify which fields you want returned by using the field parameter and listing each field separated by a comma.
		page : int
			the number of the page you would request
		page_size : int
			the number of result in the same page
		order_by: str
			By default, the result in the delivery fees results is ordered by wilaya_id in an ascending way
		desc: bool
			Order the result descening
		asc: bool
			Order the result ascending
	"""
	wilaya_id: Optional[str] = None
	fields: Optional[str] = None
	page: Optional[int] = None
	page_size: Optional[int] = None
	order_by: Optional[str] = None
	desc: Optional[bool] = None
	asc: Optional[bool] = None